#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <process.h>
#include <math.h>

#include "forkpty.h"
#include "cpu.h"

#include "terminal.h"
#include "keyboard.h"
#include "beeper.h"

#include "256color.h"

static const unsigned char p15font[15*256] = {
#include "8x15.inc"
};

static termwindow* cur_terminal = NULL;

void AddBuffer(char value)
{
    if(cur_terminal)
        cur_terminal->EchoBack((const char*)&value, 1);
}
void AddBuffer(const char* zstring)
{
    if(cur_terminal)
        cur_terminal->EchoBack(zstring, strlen(zstring));
}

static unsigned char SavedFont[256*32];
static void SaveFont()
{
    MakeFontMemAvail();
    _fmemcpy(
        SavedFont,
        MK_FP(0xA000,0x0000),
        sizeof(SavedFont));
    MakeFontMemUnavail();
}
static void RestoreFont()
{
    MakeFontMemAvail();
    _fmemcpy(
        MK_FP(0xA000,0x0000),
        SavedFont,
        sizeof(SavedFont));
    MakeFontMemUnavail();
}

static void SetScanlineCount(unsigned n)
{
    /**/
    unsigned hibit = ((n-1) & 0xFF) << 8;
    outport(0x3D4, 0x2E11);
    //outport(0x3D4, hibit | 0x06);
    outport(0x3D4, hibit | 0x12);
    outport(0x3D4, hibit | 0x15);
    outport(0x3D4, 0xAE11);
    //Default parameters for 80x50 mode listed below:
    // 6 = vertical total          -- 0xBF (8 bits, 9th is at @7 bit 5)
    // 9 = maximum scan line       -- 0x47 (5 bits = per row)
    //10 = vertical retrace start  -- 0x9C (8 bits; 9th is at @7 bit 7)
    //11 = vertical retrace end    -- 0x8E (low 4 bits)
    //12 = vertical display end    -- 0x8F (8 bits, 9th is at @7 bit 6)
    //15 = start vertical blank    -- 0x8F (8 bits, 9th is at @7 bit 3)
    //16 = end vertical blank      -- 0x8E (7 bits)
    /*
    outportb(0x3D4, 6); int val6 = inportb(0x3D5);
    outportb(0x3D4, 9); int val9 =inportb(0x3D5);
    outportb(0x3D4,0x10); int val10= inportb(0x3D5);
    outportb(0x3D4,0x11); int val11 = inportb(0x3D5);
    outportb(0x3D4,0x12); int val12 = inportb(0x3D5);
    outportb(0x3D4,0x15); int val15 = inportb(0x3D5);
    outportb(0x3D4,0x16); int val16 = inportb(0x3D5);
    fprintf(stderr, "6=%X, 9=%X, 10=%X, 11=%X, 12=%X, 15=%X, 16=%X\n",
                    val6,val9,val10,val11,val12);
    exit(0);
    */
}

void VgaSetCustomMode(
    unsigned width,
    unsigned height,
    unsigned font_height,
    int is_9pix,
    int is_half/*horizontally doubled*/,
    int is_double/*vertically doubled*/)
{
    unsigned hdispend = width;
    unsigned vdispend = height*font_height;
    if(is_double) vdispend *= 2;
    unsigned htotal = width*5/4;
    unsigned vtotal = vdispend+45;

    *(unsigned char*)MK_FP(0x40, 0x85) = font_height;

    if(font_height == 16 || font_height == 15)
    {
        // Set standard 80x25 mode as a baseline
        // This triggers font reset on JAINPUT.
        *(unsigned char*)MK_FP(0x40,0x87) |= 0x80; // tell BIOS to not clear VRAM
        _asm { mov ax, 3; int 0x10 }
        //VgaSetMode(3);
        *(unsigned char*)MK_FP(0x40,0x87) &= ~0x80;
    }

    if(font_height ==14) { _asm { mov ax, 0x1101; mov bl, 0; int 0x10 } }
    if(font_height == 8) {
        _asm { mov ax, 0x1102; mov bl, 0; int 0x10 }
    }

    if(font_height == 15) {
        _asm {
            push es
            push bp
             mov ax, seg p15font
             mov es, ax
             mov bp, offset p15font
             mov ax, 0x1100
             mov bx, 0x0F00
             mov cx, 256
             mov dx, 0
             int 0x10
            pop bp
            pop es
        }
    }

    /* This script is, for the most part, copied from DOSBox. */

    {unsigned char Seq[5] = { 0, !is_9pix + 8*is_half, 3, 0, 6 };
    for(unsigned a=0; a<5; ++a) outport(0x3C4, a | (Seq[a] << 8));}

    // Disable write protection
    outportb(0x3D4, 0x11); outportb(0x3D5, inportb(0x3D5)&0x7F);
    // crtc
    {for(unsigned a=0; a<=0x18; ++a) outport(0x3D4, a | 0x0000);}
    unsigned overflow = 0, max_scanline = 0, ver_overflow = 0, hor_overflow = 0;
    outport(0x3D4, 0x00 | ((htotal-5) << 8));   hor_overflow |= ((htotal-5) >> 8) << 0;
    outport(0x3D4, 0x01 | ((hdispend-1) << 8)); hor_overflow |= ((hdispend-1) >> 8) << 1;
    outport(0x3D4, 0x02 | ((hdispend) << 8));   hor_overflow |= ((hdispend) >> 8) << 2;
    unsigned blank_end = (htotal-2) & 0x7F;
    outport(0x3D4, 0x03 | ((0x80|blank_end) << 8));
    unsigned ret_start = is_half ? hdispend+3 : hdispend+5;
    outport(0x3D4, 0x04 | (ret_start << 8));    hor_overflow |= ((ret_start) >> 8) << 4;
    unsigned ret_end = is_half ? ((htotal-18)&0x1F)|(is_double?0x20:0) : ((htotal-4)&0x1F);
    outport(0x3D4, 0x05 | ((ret_end | ((blank_end & 0x20) << 2)) << 8));
    outport(0x3D4, 0x06 | ((vtotal-2) << 8));
    overflow |= ((vtotal-2) & 0x100) >> 8;
    overflow |= ((vtotal-2) & 0x200) >> 4;
    ver_overflow |= (((vtotal-2) & 0x400) >> 10);
    unsigned vretrace = vdispend + (4800/vdispend + (vdispend==350?24:0));
    outport(0x3D4, 0x10 | (vretrace << 8));
    overflow |= (vretrace & 0x100) >> 6;
    overflow |= (vretrace & 0x200) >> 2;
    ver_overflow |= (vretrace & 0x400) >> 6;
    outport(0x3D4, 0x11 | (((vretrace+2) & 0xF) << 8));
    outport(0x3D4, 0x12 | ((vdispend-1) << 8));
    overflow |= ((vdispend-1) & 0x100) >> 7;
    overflow |= ((vdispend-1) & 0x200) >> 3;
    ver_overflow |= ((vdispend-1) & 0x400) >> 9;
    unsigned vblank_trim = vdispend / 66;
    outport(0x3D4, 0x15 | ((vdispend+vblank_trim) << 8));
    overflow |= ((vdispend+vblank_trim) & 0x100) >> 5;
    max_scanline |= ((vdispend+vblank_trim) & 0x200) >> 4;
    ver_overflow |= ((vdispend+vblank_trim) & 0x400) >> 8;
    outport(0x3D4, 0x16 | ((vtotal-vblank_trim-2) << 8));
    unsigned line_compare = vtotal < 1024 ? 1023 : 2047;
    outport(0x3D4, 0x18 | (line_compare << 8));
    overflow |= (line_compare & 0x100) >> 4;
    max_scanline |= (line_compare & 0x200) >> 3;
    ver_overflow |= (line_compare & 0x400) >> 4;
    if(is_double) max_scanline |= 0x80;
    max_scanline |= font_height-1;
    unsigned underline = 0x1F; //vdispend == 350 ? 0x0F : 0x1F;
    outport(0x3D4, 0x09 | (max_scanline << 8));
    outport(0x3D4, 0x14 | (underline << 8));
    outport(0x3D4, 0x07 | (overflow << 8));
    outport(0x3D4, 0x5D | (hor_overflow << 8)); // S3 trio extension
    outport(0x3D4, 0x5E | (ver_overflow << 8)); // S3 trio extension
    unsigned offset = hdispend / 2;
    outport(0x3D4, 0x13 | (offset << 8));
    outport(0x3D4, 0x51 | (((offset & 0x300) >> 4) << 8));
    outport(0x3D4, 0x69 | (0 << 8)); // S3 trio
    outport(0x3D4, 0x5E | (ver_overflow << 8)); // (repeat for some reason)
    unsigned modecontrol = 0xA3;
    outport(0x3D4, 0x17 | (modecontrol << 8));
    // Enable write protection
    outportb(0x3D4, 0x11); outportb(0x3D5, inportb(0x3D5)|0x80);
    outport(0x3D4, 0x67 | (0 << 8)); // S3 trio

    unsigned misc_output = 0x63;/*0x02*/
    outportb(0x3C2, misc_output); // misc output

    {unsigned char Gfx[9] = { 0,0,0,0,0,0x10,0x0E,0x0F,0xFF };
    for(unsigned a=0; a<9; ++a) outport(0x3CE, a | (Gfx[a] << 8));}

    {unsigned char Att[0x15] = { 0x00,0x01,0x02,0x03,0x04,0x05,0x14,0x07,
                                 0x38,0x39,0x3A,0x3B,0x3C,0x3D,0x3E,0x3F,
                                 is_9pix*4, 0, 0x0F, is_9pix*8, 0 };
    _asm { mov dx,0x3DA; in al,dx }
    {for(unsigned a=0x0; a<0x15; ++a)
        { if(a == 0x11) continue;
          outportb(0x3C0, a);
          outportb(0x3C0, Att[a]); }} }
    _asm { mov dx,0x3DA; in al,dx }
    outportb(0x3C0, 0x20);

    *(unsigned char*)MK_FP(0x40,0x10) &= ~0x30;
    *(unsigned char*)MK_FP(0x40,0x10) |= 0x20;

    *(unsigned char*)MK_FP(0x40,0x65) = 0x29;
    *(unsigned char*)MK_FP(0x40, 0x4A) = width;
    *(unsigned char*)MK_FP(0x40, 0x84) = height-1;
    *(unsigned short*)MK_FP(0x40, 0x4C) = width*height*2;

    _asm { mov ah,1; mov cx,0x0D0E; int 0x10 }  // set cursor type

    /*double clock = 28322000.0;
    if(((misc_output >> 2) & 3) == 0) clock = 25175000.0;
    clock /= (is_9pix ? 9.0 : 8.0);
    clock /= vtotal;
    clock /= htotal;
    if(is_half)   clock /= 2.0;
    VidFPS = clock;

    VidW = width;
    VidH = height;
    VidCellHeight = font_height;
    VgaGetFont();*/
}


static void ReprogramVideo()
{
/*
    const unsigned n_extralines = 1;
    _asm { mov ax,3; int 0x10 }                 // set text mode
    _asm { mov ax,0x1112; xor bx,bx; int 0x10 } // select 8x8 rom font
    _asm { mov ah,1; mov cx,0x0607; int 0x10 }  // set cursor type
    *(unsigned short* _far) 0x0040004CUL = (Rows+n_extralines)*Columns*2u; // set video page size
    SetScanlineCount(400+8*n_extralines); // ADD AN EXTRA LINE OF TEXT (increase scanlines for 400 to 408)
    // This implementation is simple enough for it to work in DOSBox.
    // A 80x51 mode is still small enough that we can fit
    // four pages to 32k of VRAM. 80*51*2*4 = 32640. 80*52*2*4 = 33280.
*/
    VgaSetCustomMode(Columns, Rows+1, USE_FONT, USE_9PIX, USE_HALF, USE_DOUBLE);
    ReconfigureVideoWidth(Columns);
}

static int npages = 2;

static void ParseVideoOptions(int argc, char** argv)
{
    int round=0;
    for(int a=1; a<argc; ++a)
    {
        const char* s = argv[a];
        while(*s == '-') ++s;
        if(*s == 'f')                  USE_FONT = atoi(s+1);
        else if(*s == 'r') { ++s; goto R; }
        else if(*s == 'c') { ++s; goto C; }
        else if(*s == 'y')
        {
            char Buf[512];
            sprintf(Buf, " CYCLES=%s", ++s); // Note: Possibility of buffer overflow

            unsigned psp = getpsp();
            sprintf( (char*) MK_FP(psp, 0x80), "%s", Buf);

             _asm { db 0xFE,0x38,0x06,0x00 }
        }
        else if(*s == 'p') { npages = atoi(++s); if(npages < 1) exit(1); }
        else if(*s >= '0' && *s <= '9') switch(round++)
        {
            case 0: C: Columns = atoi(s); break;
            case 1: R: Rows    = atoi(s)-1; break;
            default: goto Err;
        }
        else Err:
        {
            fprintf(stderr, "Unrecognized video option: %s\n", s);
            exit(1);
        }
    }
    USE_9PIX   = 0;
    USE_HALF   = 0;
    USE_DOUBLE = 0;
    int chosen = 0;
    double penalty[8];

    for(round=0; round<8; ++round)
    {
        unsigned hpix = Columns * 8;
        unsigned vpix = Rows    * USE_FONT;

        if(round&1) vpix *= 2; // USE_DOUBLE (make scanlines double-tall, double vertically)
        if(round&2) hpix *= 2; // USE_HALF (for 40x25, double horizontally)
        if(round&4) hpix = (hpix/8)*9; // USE_9PIX

        double aspect = double(hpix) / double(vpix);

        double aspect_16_9_penalty = abs(16./9. - aspect);
        double aspect_4_3_penalty  = abs(4./3. - aspect);
        double aspect_pixcount_penalty = abs(sqrt(720*double(480)) - sqrt(double(hpix)*vpix));

        double p = aspect_4_3_penalty*aspect_4_3_penalty
                 + aspect_16_9_penalty*aspect_16_9_penalty;

        penalty[round] = p + aspect_pixcount_penalty*1e-3;

        if(vpix > 900) penalty[round] += 10;
    }
    for(round=0; round<8; ++round)
        if(penalty[round] < penalty[chosen])
            chosen = round;

    USE_DOUBLE = (chosen&1) != 0;
    USE_HALF   = (chosen&2) != 0;
    USE_9PIX   = (chosen&4) != 0;
}

static unsigned char OldColumns=80, OldRows=25;
static unsigned char OldFont=16, Old9pix=1, OldHalf=0, OldDouble=0;
static void SaveOldMode()
{
    SaveFont();
    _asm {
        push bp
         mov ah, 0xF; int 0x10; mov OldColumns, ah
         //
         mov ax, 0x1130; xor bx, bx; int 0x10; mov OldRows, dl; mov OldFont, cl
         //
         mov al, 1; mov dx, 0x3C4; out dx, al
         inc dx; in al, dx; mov dl, al
         and al, 1; xor al, 1; mov Old9pix, al
         mov al, dl; shr al, 3; and al, 1; mov OldHalf, al
         //
         mov al, 9; mov dx, 0x3D4; out dx, al
         inc dx; in al, dx
         shr al, 7; mov OldDouble, al
        pop bp
    }
    if(OldRows == 0) OldRows = 25; else ++OldRows;
}

char InBuffer[4096u];
int main(int argc, char** argv)
{
    setcbrk(0);

    double cpuspeed = CPUinfo() / 1e6;

    SaveOldMode();
    ParseVideoOptions(argc, argv);
    ReprogramVideo();

    ForkPTY   * tasks = new ForkPTY[npages];
    termwindow* terms = new termwindow[npages];

    {for(int task=npages; --task >= 0; )
    {
        terms[task].SetVideoPage(task);
        char Buf[512], size[8];
        switch(npages)
        {
            case 1: strcpy(size,"One"); break;
            case 2: strcpy(size,"Two"); break;
            case 3: strcpy(size,"Three"); break;
            case 4: strcpy(size,"Four"); break;
            case 5: strcpy(size,"Five"); break;
            case 6: strcpy(size,"Six"); break;
            case 7: strcpy(size,"Seven"); break;
            case 8: strcpy(size,"Eight"); break;
            case 9: strcpy(size,"Nine"); break;
            case 10: strcpy(size,"Ten"); break;
            default: sprintf(size, "%d", npages);
        }
        sqrt(1.0);// make borland c++ link floating point formats
        sprintf(Buf, "\33c"
            "\33[1;1H\33[0;40;37m\33[2J"
            "\33[%d;1H\33[1;37;44m\33[Ktty%d\33[21;30m|\33[%d;%dH"
                    "|\33[1;37mDos\33[21;30m"
                "|\33[1;37meXit"
            "\33[1;1H"
            "\33[1;33m%s %dx%d Linux TTY%s opened (%dx%d %s font).\33[K\r\n"
            "\33[21;33mRunning on %.1f MHz CPU (measured), in 16-bit real mode on DOS 5.00\33[K\r\n"
            "\33[0;37m\r\n",
            Rows+1, task+1, Rows+1,Columns-8,
            size, Columns,Rows, npages>1?"s":"", USE_9PIX?9:8, USE_FONT,
            (USE_FONT>14 || USE_9PIX) ? "VGA" : USE_FONT>8 ? "EGA" : "CGA",
            cpuspeed);
        terms[task].Write(Buf);

        /*sprintf(Buf, "Old mode to be restored is %dx%d at %d, %d %d %d\r\n",
            OldColumns,OldRows, OldFont, Old9pix, OldHalf,OldDouble);
        terms[task].Write(Buf);*/
    }}

    /*{ char Buf[128];
      sprintf(Buf, "fd=%d, pid=%d\r\n", fd,pid);
      term.Write(Buf); }*/

    int which = 0;
    int done = 0;
    while(done < npages)
    {
        termwindow& term = terms[which];
        cur_terminal = &term;

        keyboard.CheckDOS();

        if(HeldShifts[KG_ALT] && HeldKeys[0x2D])
            done = npages; // alt+x
        if(HeldShifts[KG_ALT] && HeldKeys[0x20]) // alt+d
        {
            const unsigned SwapSize = 0x8000u;
            char* videobackup = new char[SwapSize];
            if(!videobackup)
            {
                cur_terminal->Write("\33[0mUNABLE TO ALLOCATE MEMORY FOR VIDEO SWAP\n");
                HeldKeys[0x20] = 0;
                continue;
            }
            memcpy(videobackup, (const void _far*)0xB8000000UL, SwapSize);
            Reset256Color();
            BeepOff();

            int c = 0x1E00;
            short buf[10] = {'D'|c,'O'|c,'S'|c,c,'s'|c,'h'|c,'e'|c,'l'|c,'l'|c,0x107C};
            cur_terminal->myputtext_line(0,Rows, 10, buf);

            keyboard.Uninstall();

            _asm { mov ax,0x500; int 0x10 } // switch to page 0
            printf("\33[50;1H\33[0mLaunching DOS Shell. Type EXIT to return to Linux.\r\n");
            spawnl(P_WAIT, getenv("COMSPEC"), "command", NULL);

            keyboard.Install();

            ReprogramVideo();

            memcpy((void _far*)0xB8000000UL, videobackup, SwapSize);
            delete[] videobackup;
            for(int t=0; t<npages; ++t) terms[t].Write("",0); // reposition cursor on each page
            _asm { mov ax,which;mov ah,5; int 010 } // reselect the page we were on

            continue;
        }

        if(HitAltTab)
        {
            int offset = 1;
            if(HeldShifts[KG_SHIFT]) offset = npages-1;
            which = (which+offset) % npages;
            HitAltTab = 0;
            _asm {
                mov ax,which
                mov ah,5
                int 0x10
            }
            continue;
        }

        unsigned char black  = Find256Color_DontChange(0);
        unsigned char blue   = Find256Color_DontChange(4);
        unsigned char cyan   = Find256Color_DontChange(6);
        unsigned char white  = Find256Color_DontChange(7);
        unsigned char yellow = Find256Color_DontChange(11);
        unsigned char bright = Find256Color_DontChange(15);

        if(cur_terminal)
        {
            const unsigned n=Columns-24;
            short Buffer[MaxColumns];
            static int a=30, ad=-1, b=40, bd=1, counter=0;
            unsigned bluebright  = (blue << 12) | (bright << 8);
            for(unsigned c=0; c<n; ++c)
                Buffer[c] =  (HeldKeys[c]||((c+n) < 128 ? HeldKeys[c+n] : 0)) ? (bluebright | 0x0023) :
                             (c==a ? ((blue << 12) | (yellow << 8) | 0x0001)
                             :c==b ? (bluebright | 0x0001)
                             :(blue << 12));
            if(++counter>=Columns-10) { counter=0; a+=ad; b+=bd;
                               if(a==0||a==n-1)ad=-ad;
                               if(b==0||b==n-1)bd=-bd; }
            cur_terminal->myputtext_line(6,Rows, n, Buffer);
        }

        int activity = 0;
        for(int task=0; task<npages; ++task)
        {
            if(!tasks[task].Active()) continue;

            termwindow& term = terms[task];
            if(term.OutBufferLength > 0)
            {
                unsigned tail = term.OutBufferHead - term.OutBufferLength;
                if(term.OutBufferHead < term.OutBufferLength) tail += OutBufferCapacity;
                unsigned nwrite = term.OutBufferLength;
                if(term.OutBufferHead + nwrite > OutBufferCapacity)
                    nwrite = OutBufferCapacity - term.OutBufferHead;

                unsigned error=0;
                tasks[task].Send(term.OutBuffer + tail, nwrite, error);

                if(nwrite > 0 && nwrite != 0xFFFF)
                {
                    if(nwrite > term.OutBufferLength) nwrite = term.OutBufferLength;
                    term.OutBufferLength -= nwrite;
                    term.OutBufferHead   += nwrite;
                    if(term.OutBufferHead >= OutBufferCapacity)
                        term.OutBufferHead -= OutBufferCapacity;
                    if(term.OutBufferLength == 0)
                        { term.OutBufferHead = 0; }
                    ++activity;
                }
            }

            unsigned nread = sizeof(InBuffer), error=0;
            //memcpy(InBuffer, "kupo", 4);
            tasks[task].Recv(InBuffer, nread, error);

            if(nread > 0 && nread != 0xFFFF)
            {
                term.Write( InBuffer, nread );
                ++activity;
            }
            if(error != 0)
            {
                done += 1;
                char Buf[128];
                sprintf(Buf,  "\33[0m\r\nSimulation ended in tty%d [nread=%d,errno=%u]\r\n",
                    task+1, nread, error);
                term.Write(Buf);

                tasks[task].Close();
            }
        }

        if(activity) continue;

//        wait();

        CheckBeepOff();

        if(cur_terminal)
        {
            cpuspeed = CPUinfo() / 1e6;
            char Buf[16] = {0};
            if(cpuspeed < 10.0)
                sprintf(Buf, "%8.3f MHz", cpuspeed);
            else if(cpuspeed < 100.0)
                sprintf(Buf, "%8.2f MHz", cpuspeed);
            else if(cpuspeed < 1000.0)
                sprintf(Buf, "%8.1f MHz", cpuspeed);
            else if(cpuspeed < 3000.0)
                sprintf(Buf, "%8.0f MHz", cpuspeed);
            else if(cpuspeed < 10000.0)
                sprintf(Buf, "%8.3f GHz", cpuspeed/1e3);
            else if(cpuspeed < 100000.0)
                sprintf(Buf, "%8.2f GHz", cpuspeed/1e3);
            else if(cpuspeed < 1000000.0)
                sprintf(Buf, "%8.1f GHz", cpuspeed/1e3);
            else
                sprintf(Buf, "%8.0f GHz", cpuspeed/1e3);

            cur_terminal->AdjustColors(0,Rows, 4,1, blue,white);
            cur_terminal->AdjustColors(4,Rows, 2,1, blue,black);
            cur_terminal->AdjustColors(Columns-4,Rows, 4,1, blue,white);
            cur_terminal->AdjustColors(Columns-5,Rows, 1,1, blue,black);
            cur_terminal->AdjustColors(Columns-8,Rows, 3,1, blue,white);
            cur_terminal->AdjustColors(Columns-9,Rows, 1,1, blue,black);

            unsigned attr = (blue << 12) | (cyan << 8);
            short Buf2[16];
            const unsigned n = 13;
            for(unsigned c=0; c<n; ++c)
                Buf2[c] = ((unsigned char)Buf[c]) | attr;
            cur_terminal->myputtext_line(Columns-21,Rows,n, Buf2);
        }
    }

    delete[] tasks;
    delete[] terms;

    setcbrk(1);

    BeepOff();

    //SetScanlineCount(400);
    //VgaSetCustomMode(80,25, 16, 1, 0, 0);
    /*fprintf(stderr, "Restoring mode %dx%d at %d, %d %d %d\n",
        OldColumns,OldRows, OldFont, Old9pix, OldHalf,OldDouble);*/
    VgaSetCustomMode(OldColumns,OldRows, OldFont, Old9pix, OldHalf,OldDouble);
    RestoreFont();

    unsigned pagesize_new = *(unsigned short _far*)MK_FP(0x40, 0x4C);
    unsigned blank_start = (Rows-2)*Columns*2;

    if(blank_start < pagesize_new)
        myfilltext(blank_start, (pagesize_new-blank_start) >> 1, 0x0720);

    unsigned oldc = OldColumns, curc = Columns;
    oldc <<= 1; curc <<= 1;
    unsigned rows = OldRows; if(rows > Rows) rows = Rows;

    if(oldc < curc) // going back to a smaller screen
        for(unsigned r=1; r<rows; ++r)
            mymovevidwords(r*curc, r*oldc, OldColumns);
    else if(oldc > curc) // going back to a larger screen
    {
        for(unsigned r=rows; r-- > 0; )
        {
            mymovevidwords(r*curc, r*oldc, Columns);
            myfilltext(r*oldc+curc, OldColumns-Columns, 0x0720);
        }
    }
    if(rows < OldRows)
        myfilltext(rows*OldColumns*2, (OldRows-rows)*OldColumns, 0x0720);
    Reset256Color();

    return 0;
}
